from .sylia import Clock as clock, init, screenWidth, screenHeight, screenCenter, draw
from .sylia import Sound as sound, Key as key, Mouse as mouse, running, setBackgroundColour

from .image import Image as image
from .shapes import Shape as shape
from .text import Text as text
from .duck import Duck as duck